#define UwiApi_c

// ************************************************************************
// * uwipi.c                                                              *
// *                                                                      *
// * Written by Fred Treasure and Pete Reynolds (IBM SP Linux development *
// * team)                                                                *
// * (C) Copyright IBM Corporation, 1998-2003                             *
// *                                                                      *
// *  This software may be used and distributed according to the terms    *
// *  of the Lesser GNU Public License, incorporated herein by reference. *
// ************************************************************************

// **********************************************************************
// * uwiapi.c - This source file implements the shared library for the  *
// *            IBM Service Processor adapter.                          *
// *                                                                    *
// * NOTE :                                                             *
// *    - API return codes are defined in sysdrv.h!                     *
// *    - Driver return codes are defined in uwioctl.h!                 *
// *                                                                    *
// * Exported Functions                                                 *
// * ------------------                                                 *
// * OpenSPDriverNode       - opens a specific node in the driver       *
// * OpenSPDriver           - opens the primary node                    *
// * CloseSPDriver          - closes the driver                         *
// * SystemDataIO           - sends a dot command to ASM                *
// * RegisterForReverseHB   - initiates the reverse heartbeat process   *
// * DeRegisterForReverseHB - terminates the reverse heartbeat process  *
// * RegisterForEvents      - registers app for event notification      *
// * DeregisterForEvents    - de-registers app for event notification   *
// *                                                                    *
// * Change History						                                *
// * ----------------------------------------------------------------   *
// *   Date   - Change ID  - Description                                *
// * ----------------------------------------------------------------   *
// * 11/18/02 - PHR_208637 - Initial code added to source library       *
// * ----------------------------------------------------------------   *
// * 10/01/03 - PHR_225158 - Add 5.2.16 reset event support             *
// * ----------------------------------------------------------------   *
// * 06/10/04 - T. Yu      - Fix the defect 244089 and 242835f_1        *
// * ----------------------------------------------------------------   *
// * 06/23/04 - T. Yu      - stop retries if sp reset to prevent from   *
// *                         errorcode 8 returned to lflash   
// **********************************************************************
#include "ibmasm.h"

// **********************************************************************
//              PROTOTYPES FOR LIBRARY ENTRY POINTS
// **********************************************************************
unsigned int OpenSPDriver( SYS_HANDLE    *Handle,
                           unsigned int   BfrLgth,
                           unsigned long  Reserved);


unsigned int OpenSPDriverNode(SYS_HANDLE    *Handle,
                              unsigned int   BfrLgth,
                              unsigned int   Node,
                              unsigned int  *pNumPCISPs,
                              unsigned long  Reserved);


unsigned int CloseSPDriver(SYS_HANDLE     Handle,
                           unsigned long  Reserved);


unsigned int SystemDataIO(SYS_HANDLE     Handle,
                          void          *CmdBlock,
                          unsigned long  Reserved);

unsigned int RegisterForEvents(SYS_HANDLE Handle,
                               void           *CmdBlock,
                               short           count,
                               unsigned char **CmdArray,
                               unsigned long   Reserved);

unsigned int DeregisterForEvents(SYS_HANDLE    Handle,
                                 unsigned long Reserved);

unsigned int RegisterForReverseHB(SYS_HANDLE    Handle,
                                  unsigned long reserved);


unsigned int DeRegisterForReverseHB(SYS_HANDLE    Handle,
                                    unsigned long reserved);


// **********************************************************************
//              LOCAL USB WRAPPER PROTOTYPES                            *
// **********************************************************************
unsigned int OpenUsbSPDriverNode( SYS_HANDLE    *Handle,
                                  unsigned int   BfrLgth,
                                  unsigned int   Node,
                                  unsigned int  *pNumPCISPs,
                                  unsigned long  Reserved);


// **********************************************************************
//                  LOCAL FUNCTION PROTOTYPES
// **********************************************************************
int AddCaller( PSP_SHARED_MEM    pSpSharedMem,
               int               CallersPid,
               unsigned int      BuffSize,
               unsigned int      Node,
               int               Handle);

int InitDeviceNodes(int SpCount);

int InitSharedMem(PSP_SHARED_MEM pSpSharedMem);

int RemoveCaller(PSP_SHARED_MEM pSpSharedMem, int handle, int CallersPid);

int ValidateCaller(PSP_SHARED_MEM pSpSharedMem, int Handle, int CallersPid);

int GetHandle(PSP_SHARED_MEM pSpSharedMem);

int FreeHandle(PSP_SHARED_MEM pSpSharedMem, int Handle);

int GetNodeFromHandle(PSP_SHARED_MEM pSpSharedMem, SYS_HANDLE Handle);

int LockSharedMemory(void);

void UnLockSharedMemory(int FileDesc);

unsigned char  usb_cksum(unsigned char* buf, int size);

void InitResetEventBuff(PSP_SHARED_MEM pSpSharedMem, unsigned char *EventBuff);

// ************************************************************************
// * OpenSPDriver - This routine exists to support pre multi node systems.*
// *                It is simply a pass thru to OpenSPDriverNode, passing *
// *                OpenSPDriverNode a default node of 0 and a NULL ptr   *
// *                to the number of SPs available.                       *
// *                                                                      *
// * Valid Return Codes to Application :                                  *
// *      RC_ALREADY_OPEN         - driver already opened - can't re-open *
// *      RC_INVALID_HANDLE       - invalid driver handle                 *
// *      RC_INVALID_BUFFER       - illegal buffer length                 *
// *      RC_SYSTEM_ERROR         - semaphore error or failed to open     *
// *      RC_SUCCESS              - close was successful                  *
// ************************************************************************
unsigned int OpenSPDriver( SYS_HANDLE   *Handle,
                           unsigned int  IOBufferLength,
                           unsigned long Reserved)
{
     return OpenUsbSPDriverNode(Handle, IOBufferLength, 0, NULL, Reserved);
}



// ************************************************************************
// * OpenSPDriverNode - wrapper call to OpenUsbSPDriver()                 *
// *                                                                      *
// * Valid Return Codes to Application :                                  *
// *      RC_ALREADY_OPEN         - driver already opened - can't re-open *
// *      RC_INVALID_BUFFER       - illegal buffer length                 *
// *      RC_SYSTEM_ERROR         - failed to open or ioctl or malloc     *
// *      RC_SUCCESS              - close was successful                  *
// ************************************************************************
unsigned int  OpenSPDriverNode( SYS_HANDLE         *Handle,
                                unsigned int        IOBufferLength,
                                unsigned int        Node,
                                unsigned int       *pNumMultiNodeSPs,
                                unsigned long       Reserved)
{
    return OpenUsbSPDriverNode(Handle, IOBufferLength, Node, pNumMultiNodeSPs, Reserved);
}





// ************************************************************************
// * OpenUsbSPDriverNode - This routine must be called first by a app     *
// *                    before any other API call can be made. This opens *
// *                    a specific node in the driver, registers the app's*
// *                    PID with the driver                               *
// *                                                                      *
// * Valid Return Codes to Application :                                  *
// *      RC_ALREADY_OPEN         - driver already opened - can't re-open *
// *      RC_INVALID_BUFFER       - illegal buffer length                 *
// *      RC_SYSTEM_ERROR         - failed to open or ioctl or malloc     *
// *      RC_SUCCESS              - close was successful                  *
// ************************************************************************
unsigned int  OpenUsbSPDriverNode( SYS_HANDLE     *Handle,
                                   unsigned int    IOBufferLength,
                                   unsigned int    Node,
                                   unsigned int   *pNumMultiNodeSPs,
                                   unsigned long   Reserved)
{
   int i;
   int rc;
   int CallersPid;
   int Timeout;
   int LockFileDesc;
   int LocalHandle;


   int MemID;
   PSP_SHARED_MEM pSpSharedMem;
   void *SharedMem;

   if(debug) syslog(LOG_CRIT, "Entering OpenSPDriverNode for node %1.\n", Node);
   // ENSURE THAT THE SP BUFFER LENGTH IS VALID (1 - 32 Kb)
   if ( (IOBufferLength < 1)  ||  (IOBufferLength > 32) )
   {
       if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode: Invalid buffer length passed by caller, returning RC_INVALID_BUFFER.\n");
       return RC_INVALID_BUFFER;
   }

   // ALLOCATE SHARED MEMORY
   pSpSharedMem = NULL;

   MemID = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

   if ( MemID == IPC_FAILURE )
   {
       if(debug) syslog(LOG_CRIT, "OpenSPDriverNode shmget() failed, returning RC_SYSTEM_ERROR.\n");
       return RC_SYSTEM_ERROR;
   }

   // ATTACH THIS PROCESS TO THE SHARED MEMORY
   SharedMem = shmat(MemID, (void *)0, 0);

   if (SharedMem == (void *)IPC_FAILURE )
   {
       if(debug) syslog(LOG_CRIT, "OpenSPDriverNode shmat() failed, returning RC_SYSTEM_ERROR.\n");
       return RC_SYSTEM_ERROR;
   }

   // SET UP THE LOCAL POINTER TO SHARED MEMORY
   pSpSharedMem = (PSP_SHARED_MEM)SharedMem;

   // ENSURE THAT CALLERS THE POINTER TO HANDLE IS A VALID NON-NULL POINTER
   if( Handle == NULL )
   {
      // FILL IN THE CALLERS SP COUNT
      if (pNumMultiNodeSPs)
      {
          if(debug) syslog(LOG_CRIT, "OpenSPDriverNode returning SpCount, SpCount = %i.\n", pSpSharedMem->SpCount);

          *pNumMultiNodeSPs = pSpSharedMem->SpCount;
          rc = RC_SUCCESS;
      }
      else
      {
         if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode Invalid pointer to NumMultiNodes passed by caller, returning RC_INVALID_HANDLE.\n");
         rc = RC_INVALID_HANDLE;
      }

      shmdt(SharedMem);
      return rc;
   }

   // MAKE SURE THAT THE SP IS NOT CURRENTLY RESETTING AND CHECK
   // FOR AT LEAST 1 VALID SP IN THE SYSTEM
   Timeout = 0;

   while ((pSpSharedMem->SpCount == 0) || (pSpSharedMem->Resetting))
   {
       sleep(1);

       Timeout++;

       if (Timeout >= CMD_TIMEOUT)
       {
		   if (pSpSharedMem->SpCount == 0)
		   {
              if(debug) syslog(LOG_CRIT, "OpenSpDriverNode() Could not find a valid SP device in the system, returning RC_OPEN_FAILED.\n");
           }
		   else
		   {
		      if(debug) syslog(LOG_CRIT, "OpenSpDriverNode() Detected SP reset in progress, returning RC_OPEN_FAILED.\n");
           }

           shmdt(SharedMem);
	       return RC_OPEN_FAILED;
       }
   }

   // ENSURE THE CALLERS NODE IS VALID
   if ( (Node >= MAX_SP_NODES) || (Node < 0) )
   {
       if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode Detected an invalid node, node = %i, returning RC_OPEN_FAILED.\n", Node);
       shmdt(SharedMem);
       return RC_OPEN_FAILED;
   }

   // ENSURE THE NODE IS OPEN AND ACCESSABLE
   while(pSpSharedMem->SPDevNode[Node].State != NODE_OPENED)
   {
       sleep(2);

       Timeout++;

       if (Timeout >= CMD_TIMEOUT)
       {
	       if(debug) syslog(LOG_CRIT, "OpenSPDriverNode detected an invalid node state, node= %i  State = %i.\n",Node, pSpSharedMem->SPDevNode[Node].State);
           shmdt(SharedMem);
           return RC_OPEN_FAILED;
       }
   }


   LockFileDesc = LockSharedMemory();

   if (LockFileDesc == -1)
   {
       if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode could not get shared memory access, returning RC_TIEMOUT.\n");
       shmdt(SharedMem);
       return RC_OPEN_FAILED;
   }

   if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode locked shared memory.\n");


   CallersPid = getpid();

   if(debug) syslog(LOG_CRIT, "OpenSpDriverNode callers pid = %i. \n",CallersPid);

   // ENSURE THAT THIS PROCESS ONLY CALLS OPEN ONCE.
   for (i = 1; i <= MAX_HANDLES; i++)
   {
       if ((pSpSharedMem->CallerList[i].Pid == CallersPid) && (pSpSharedMem->CallerList[i].Node == (int)Node))
       {
           if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode: PID already has an open handle to the driver, returning RC_ALREADY_OPEN.\n");
           shmdt(SharedMem);
	       UnLockSharedMemory(LockFileDesc);
           if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode Already Open Exit - unlocked shared memory.\n");
           return RC_ALREADY_OPEN;
       }
   }

   // GET A HANDLE
   LocalHandle = GetHandle(pSpSharedMem);

   if (LocalHandle == INVALID_USB_HANDLE)
   {
      if(debug) syslog(LOG_CRIT, "OpenSPDriverNode could not get a handle, returning RC_OPEN_FAILED.\n");

	  shmdt(SharedMem);
      UnLockSharedMemory(LockFileDesc);
      if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode unlocked shared memory - 1.\n");
      return RC_OPEN_FAILED;
   }

   // ADD THIS CALLER TO THE LIST OF CALLERS
   rc = AddCaller(pSpSharedMem, CallersPid, IOBufferLength, Node, LocalHandle);

   if (rc != RC_SUCCESS)
   {
       // CLEAN UP AND EXIT
       if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode AddCaller() could not add this caller to the caller list, returning RC_OPEN_FAILED.\n");
       RemoveCaller(pSpSharedMem, LocalHandle, CallersPid);
       shmdt(SharedMem);
       UnLockSharedMemory(LockFileDesc);
       if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode unlocked shared memory - 2.\n");
       return RC_OPEN_FAILED;
   }


   // FILL IN THE CALLERS SP COUNT AND HANDLE
   if (pNumMultiNodeSPs )
   {
       *pNumMultiNodeSPs = pSpSharedMem->SpCount;
   }

   *Handle = LocalHandle;

   // UNLOCK AND DETATCH SHARED MEMORY

   if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode successfully opened Service Processor USB node %i.\n", Node);
   shmdt(SharedMem);
   UnLockSharedMemory(LockFileDesc);
   if(debug) syslog(LOG_CRIT, "OpenUsbSPDriverNode unlocked shared memory - 3.\n");
   if(debug) syslog(LOG_CRIT, "Exiting OpenUsbSPDriverNode\n");
   return RC_SUCCESS;
}


// ************************************************************************
// ************************************************************************
unsigned int SystemDataIO( SYS_HANDLE Handle, void *CmdBlock, unsigned long  Reserved)
{
    int SleepTime;
    int rc;
    int BuffLen;
    int CallersPid;
    int Node;
    int Timeout;
    int LockFileDesc;
	int Finished;
	int RetryCount;

    unsigned char *TempPtr;
    unsigned char  checksum;

    // shared mem vars
    int   MemID;
    void *SharedMem = (void *)0;
    PSP_SHARED_MEM pSpSharedMem;

    // DISPLAY WELCOME MESSAGE
    //printf("ASMLIB: SystemDataIO called to handle a command request. pid = %i\n", getpid());

    // VALIDATE PARAMETERS
    if (CmdBlock == NULL)
    {
       if (debug) printf("ASMLIB: Invalid buffer pointer received by caller, returning RC_INVALID_BUFFER.\n");
       return RC_INVALID_BUFFER;
    }

    if (Handle <= 0 || Handle > MAX_HANDLES)
    {
       if (debug) printf("ASMLIB: Invalid handle received, returning RC_INVALID_HANDLE.\n");
       return RC_INVALID_HANDLE;
    }

    // ALLOCATE SHARED MEMORY
	pSpSharedMem = NULL;

	MemID = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

	if (MemID == IPC_FAILURE)
	{
		 //if(debug) syslog(LOG_CRIT, "SystemDataIO: shmget() failed, returning RC_SYSTEM_ERROR.\n");
		 return RC_SYSTEM_ERROR;
	}

	// ATTACH THIS PROCESS TO THE SHARED MEMORY
	SharedMem = shmat(MemID, (void *)0, 0);

	if (SharedMem == (void *)IPC_FAILURE )
	{
		 //if(debug) syslog(LOG_CRIT, "SystemDataIO: shmat() failed, returning RC_SYSTEM_ERROR.\n");
		 return RC_SYSTEM_ERROR;
	}

    // SET UP THE LOCAL POINTER TO SHARED MEMORY
	pSpSharedMem = (PSP_SHARED_MEM)SharedMem;

	RetryCount = 0;
	Finished = FALSE;

	while(!Finished)
	{
		// MAKE SURE THAT THE SP IS NOT CURRENTLY RESETTING AND CHECK
		// FOR AT LEAST 1 VALID SP IN THE SYSTEM
		Timeout = 0;

		while ((pSpSharedMem->SpCount == 0) || (pSpSharedMem->Resetting))
		{
			sleep(1);
			if(debug) syslog(LOG_CRIT, "SystemDataIO: waiting for reset to complete (Timeout = %d)\n", Timeout);
			Timeout++;
			
			if (Timeout >= CMD_TIMEOUT_RESET)
			{
				shmdt(SharedMem);
				printf ("SystemDataIO: time out\n");
				return RC_TIMEOUT;
			}
		}

		LockFileDesc = LockSharedMemory();

		if (LockFileDesc == -1)
		{
			//if(debug) syslog(LOG_CRIT, "ASMLIB:SystemDataIO: Could not get access to shared memory, Returning RC_TIMEOUT.\n");
			shmdt(SharedMem);
			return RC_TIMEOUT;
		}

		//if(debug) syslog(LOG_CRIT, "SystemDataIO() Locked shared memory.\n");

		Node = GetNodeFromHandle(pSpSharedMem, Handle);

		// ENSURE THE CALLERS NODE IS VALID
		if ( (Node >= MAX_SP_NODES) || (Node < 0) ) 	
		{
			syslog(LOG_CRIT, "ASMLIB: detected an invalid node, returning RC_INVALID_HANDLE.\n");
			shmdt(SharedMem);
			UnLockSharedMemory(LockFileDesc);
            //if(debug) syslog(LOG_CRIT, "SystemDataIO() unlocked shared memory.\n");
			return RC_INVALID_HANDLE;
		}

		// ENSURE THE NODE IS OPEN AND ACCESSABLE
		if(pSpSharedMem->SPDevNode[Node].State != NODE_OPENED)
		{
			//if(debug) syslog(LOG_CRIT, "ASMLIB: SystemDataIO: node [%d] not open yet!  detected an invalid node, returning RC_SYSTEM_ERROR. \n", Node);
			shmdt(SharedMem);
			UnLockSharedMemory(LockFileDesc);
            //if(debug) syslog(LOG_CRIT, "SystemDataIO() unlocked shared memory.\n");
			return RC_SYSTEM_ERROR;
		}

		// VALIDATE CALLERS INFO
		CallersPid = getpid();

		rc = ValidateCaller(pSpSharedMem, Handle, CallersPid);

		if (rc != RC_SUCCESS)
		{
			syslog(LOG_CRIT, "ASMLIB: ValidateCaller() failed, returning RC_INVALID_HANDLE.\n");
			shmdt(SharedMem);
			UnLockSharedMemory(LockFileDesc);
            //if(debug) syslog(LOG_CRIT, "SystemDataIO() unlocked shared memory.\n");
			return RC_SYSTEM_ERROR;
		}
		else
		{
			pSpSharedMem->CallerList[Handle].Pid = CallersPid;
		}

		// INITIALIZE THE SHARED MEMORY BUFFER
		memset(pSpSharedMem->Buffer, 0, MAX_BUFF_SIZE );

		// CALCULATE THE LENGTH OF THE DATA TO BE SENT TO THE CYPRESS CHIP
		// WHICH INCLUDES THE LENGTH OF THE CALLERS COMMAND PACKET PLUS AN
		// EXTRA BYTE OF CHECKSUM DATA
		BuffLen =  CHK_SUM_LEN                    +          // 1 byte of check sum
				   CMD_BLK_LEN                    +          // 6 bytes of command data
				   ((unsigned char *)CmdBlock)[1] +          // command length
				   ((unsigned char *)CmdBlock)[2] +          // data length LSB
				   ((unsigned char *)CmdBlock)[3] * 256;     // data length MSB

		if ( BuffLen > MAX_BUFF_SIZE )
		{
			//if(debug) syslog(LOG_CRIT, "systemDataIO: Invalid buffer length sent by pid %i, returning EXIT_FAILURE.\n",pSpSharedMem->CallersPid);
			shmdt(SharedMem);
			UnLockSharedMemory(LockFileDesc);
            //if(debug) syslog(LOG_CRIT, "SystemDataIO() unlocked shared memory.\n");
			return RC_INVALID_BUFFER;
		}

		pSpSharedMem->BuffLen = BuffLen;

		// GET THE CHECKSUM
		checksum = usb_cksum( CmdBlock, BuffLen - CHK_SUM_LEN );

		// FILL IN THE FIRST 2 BYTES OF THE BUFFER WITH THE BUFFER LENGTH
		pSpSharedMem->Buffer[0] = (unsigned char)((BuffLen & 0xFF00)>>8 );
		pSpSharedMem->Buffer[1] = (unsigned char)( BuffLen & 0xFF);

		TempPtr = pSpSharedMem->Buffer;

		// GET THE TEMP PTR BEYOND THE BUFFER LENGTH SO WE CAN COPY THE CALLERS
		// DATA INTO THE SHARED BUFFER
		TempPtr++;
		TempPtr++;

		// NOW COPY THE CALLERS COMMAND PACKET INTO THE SHARED MEMORYS BUFFER
		memcpy(TempPtr, CmdBlock, BuffLen - CHK_SUM_LEN);

		// PUT THE CHECKSUM AT THE END OF THE CALLERS DATA
		// ADD 1 TO COMPENSATE FOR THE 2 BYTES OF DATA LENGTH PRECEEDING
		// THE CALLERS DATA AND THE CHK SUM LENGTH AT THE END OF THE DATA
		pSpSharedMem->Buffer[BuffLen + 1] = checksum;

		// UNLOCK SHARED MEMORY BEFORE WE SIGANL IBMASM TO HANDLE THE COMMAND
		//unlink(MEM_LOCK_FILE);

		// SIGNAL IBMASM THAT THERE IS A COMMAND TO PROCESS
		kill(pSpSharedMem->IbmasmsPid, SIGUSR1);

		Timeout = 0;

		while(!pSpSharedMem->CmdRespReceived)
		{
			usleep(20000);
			Timeout+=1;

			if (Timeout >= CMD_TIMEOUT * 500)
			{
				shmdt(SharedMem);
				UnLockSharedMemory(LockFileDesc);
                //if(debug) syslog(LOG_CRIT, "SystemDataIO() unlocked shared memory.\n");
				return RC_TIMEOUT;
			}
		}

		pSpSharedMem->CmdRespReceived = FALSE;

		if (pSpSharedMem->CallerList[Handle].ReturnCode == RC_SUCCESS)
	    {
            rc = pSpSharedMem->CallerList[Handle].ReturnCode;
		    Finished = TRUE;
		    break;
		}
		else
		{
            RetryCount++;

			if ((pSpSharedMem->CallerList[Handle].ReturnCode == RC_WRITE_ERROR) ||
			    (pSpSharedMem->CallerList[Handle].ReturnCode == RC_READ_ERROR) )
			{
            	UnLockSharedMemory(LockFileDesc);
                //if(debug) syslog(LOG_CRIT, "SystemDataIO(): RC_WRITE_ERRROR FOUND, SLEEP 5 SEC\n");
	            sleep(5);
			}

			//if(debug) syslog(LOG_CRIT, "SystemDataIO: Shared Object detected a failed command, retry count = %i.\n", RetryCount);
		
			if (RetryCount >= MAX_RETRYS)
			{
				//if(debug) syslog(LOG_CRIT, "systemDataIO: Shared Object setting reset flag due to failed commands.\n");
				pSpSharedMem->Resetting = TRUE;
				shmdt(SharedMem);
                UnLockSharedMemory(LockFileDesc);
                
                //if(debug) syslog(LOG_CRIT, "SystemDataIO() unlocked shared memory.\n");
				return RC_TIMEOUT;
			}
		
			
        }
	}

    // ENSURE CALLERS BUFFER IS LARGE ENOUGH TO HOLD THE COMMAND RESPONSE
    if (pSpSharedMem->BuffLen > pSpSharedMem->CallerList[Handle].BufferSize)
    {
        //printf("ASMLIB: Command Response to large to fit into callers buffer, returning RC_INVALID_BUFFER.\n");
        BuffLen = pSpSharedMem->CallerList[Handle].BufferSize;
        rc = RC_INVALID_BUFFER;
    }
    else
    {
        BuffLen = pSpSharedMem->BuffLen;
        rc = RC_SUCCESS;
    }

    TempPtr = pSpSharedMem->Buffer;

    // NEED TO GET THE SRC BUFFER POINTER BEYOND THE USB BUFFER LENGTH
    TempPtr++;
    TempPtr++;

    // COPY THE COMMAND RESPONSE FROM THE SHARED MEMORY BUFFER TO THE
    // CALLERS BUFFER
    memcpy( CmdBlock, TempPtr, BuffLen);

    pSpSharedMem->CallersNode   = -1;
    pSpSharedMem->CallersPid    = -1;
    pSpSharedMem->CallersHandle = -1;

    // CLEAN UP AND EXIT
	shmdt(SharedMem);
    UnLockSharedMemory(LockFileDesc);
    //if(debug) syslog(LOG_CRIT, "SystemDataIO() unlocked shared memory.\n");

    // SLEEP JUST A BIT BEFORE RETURNING TO THE CALLER, MAKING SURE
    // OTHER TASKS HAVE TIME TO RUN, NOT SURE WE NEED THIS THOUGH
    // THE FOLLOWING RANDOM TIME GENERATOR WAS TAKEN FROM THE
    // LINUX MAN PAGES FOR RAND()
    SleepTime = 1+ (int)(10.0*rand()/(RAND_MAX+1.0));
    usleep(SleepTime);

    //printf("ASMLIB: Waiting for %i uSeconds before returning to caller.\n",
    //        SleepTime);

    return rc;
}


// ************************************************************************
// ************************************************************************
unsigned int RegisterForEvents(SYS_HANDLE      Handle,
                               void           *CmdBlock,
                               short int       count,
                               unsigned char **CmdArray,
                               unsigned long   Reserved)
{
    int i;
    int rc;
    int EvtCmdLen;
    int EventIndex;
    int IgnoreEvent;
    int Timeout;
    int Node;
    int LockFileDesc;
    int BuffLen = 0;

    unsigned char EventBuff[50];

    unsigned char *TempPtr;

    // SHARED MEMORY DECLARATIONS
    int   MemID;
    void *SharedMem;

    PSP_SHARED_MEM pSpSharedMem;

    // VALIDATE CALLERS PARMS
    if ( (Handle <= 0)  || (Handle > MAX_HANDLES) )
    {
        //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Invalid Handle detected, returning RC_INVALID_HANDLE.\n");
        return RC_INVALID_HANDLE;
    }

    if (CmdBlock == NULL)
    {
        //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Invalid command block detected, returning RC_INVALID_BUFFER.\n");
        return RC_INVALID_BUFFER;
    }

    // ENSURE THAT THE IGNORE LIST CONTAINS ALL VALID STRINGS
    if (count)
    {
        if (CmdArray == NULL)
        {
            //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Invalid command array detected, returning RC_INVALID_BUFFER.\n");
	        return RC_INVALID_BUFFER;
        }

        for ( i = 0; i < count; i++)
        {
           if (CmdArray[i] == NULL)
           {
               //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Invalid command array detected, returning RC_INVALID_BUFFER.\n");
               return RC_INVALID_BUFFER;
           }
        }
    }

	// ALLOCATE SHARED MEMORY
	pSpSharedMem = NULL;

	MemID = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

    if (MemID == IPC_FAILURE)
	{
		 //if(debug) syslog(LOG_CRIT, "RegisterForEvents() shmget() failed, returning RC_SYSTEM_ERROR.\n");
		 return RC_SYSTEM_ERROR;
	}

	// ATTACH THIS PROCESS TO THE SHARED MEMORY
	SharedMem = shmat(MemID, (void *)0, 0);

	if (SharedMem == (void *)IPC_FAILURE )
	{
		 //if(debug) syslog(LOG_CRIT, "RegisterForEvents() shmat() failed, returning RC_SYSTEM_ERROR.\n");
		 return RC_SYSTEM_ERROR;
	}

	// SET UP THE LOCAL POINTER TO SHARED MEMORY
	pSpSharedMem = (PSP_SHARED_MEM)SharedMem;

    // MAKE SURE THAT THE SP IS NOT CURRENTLY RESETTING AND CHECK
	// FOR AT LEAST 1 VALID SP IN THE SYSTEM
	Timeout = 0;

	while ((pSpSharedMem->SpCount == 0) || (pSpSharedMem->Resetting))
	{
		sleep(1);

		Timeout++;

		if (Timeout >= CMD_TIMEOUT)
		{
			//if (pSpSharedMem->SpCount == 0)
			//{
				 //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Could not find a valid SP device in the system, returning RC_TIMEOUT.\n");
			//}
			//else
			//{
			//if(debug) syslog(LOG_CRIT, "RegisterForEvents() Detected SP reset in progress, returning RC_TIMEOUT.\n");
			//}

			shmdt(SharedMem);
			return RC_TIMEOUT;
		}
	}

    // ENSURE HANDLE IS VALID
    if (pSpSharedMem->HandleList[Handle].Free)
    {
        //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Invalid Handle detected, returning RC_INVALID_HANDLE.\n");
		shmdt(SharedMem);
        return RC_INVALID_HANDLE;
    }

    Node = GetNodeFromHandle(pSpSharedMem, Handle);

    // ENSURE THE CALLERS NODE IS VALID
    if ( (Node >= MAX_SP_NODES) || (Node < 0) )
    {
        //if(debug) syslog(LOG_CRIT, "RegisterForEvents() detected an invalid node, returning RC_OPEN_FAILED.\n");
		shmdt(SharedMem);
        return RC_INVALID_HANDLE;
    }

    // AN AVAILABLE NODE MEANS THAT THE NODE WAS NEVER INTIALIZED
    if(pSpSharedMem->SPDevNode[Node].State == NODE_AVAILABLE)
    {
        //if(debug) syslog(LOG_CRIT, "RegisterForEvents() detected an invalid node, returning RC_OPEN_FAILED.\n");
		shmdt(SharedMem);
        return RC_SYSTEM_ERROR;
    }

    Timeout = 0;

    // WE MAY HAVE TO WAIT ON A RESET TO COMPLETE
    while (pSpSharedMem->SPDevNode[Node].State != NODE_OPENED)
    {
		sleep(1);

		Timeout++;

		if (Timeout >= CMD_TIMEOUT)
		{
			shmdt(SharedMem);
            return RC_TIMEOUT;
        }
	}

    LockFileDesc = LockSharedMemory();

    if (LockFileDesc == -1)
    {
         //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Could not get access to shared memory, Returning RC_OPEN_FAILED.\n");
	     shmdt(SharedMem);
         return RC_SYSTEM_ERROR;
    }

    //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Locked shared memory.\n");


    // IF THIS PROCESS IS ALREADY INITIALIZED, DON'T RE-INITIALIZE IT
    if (!pSpSharedMem->CallerList[Handle].EventRegInfo.Initialized)
    {
        //if(debug) syslog(LOG_CRIT, "RegisterForEvents() New registration initializing event structure.\n");

        // UPDATE SHARED MEMORY AND WAIT UNTIL AN EVENT OCCURS
        pSpSharedMem->CallerList[Handle].EventRegInfo.Initialized = TRUE;
        pSpSharedMem->CallerList[Handle].EventRegInfo.Pid         = getpid();
        pSpSharedMem->CallerList[Handle].EventRegInfo.Handle      = Handle;
        pSpSharedMem->CallerList[Handle].EventRegInfo.Node        = Node;
        pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex  = pSpSharedMem->SPDevNode[Node].EventIndex;

	    pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting       = TRUE;
	    pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived = FALSE;

        //printf("ASMLIB: Setting the callers event index to %i.\n",
        //        pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex );

    }
    else
    {
        //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Process already registered, returning RC_ALREADY_REGISTERED.\n");

        if (pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting)
        {
			shmdt(SharedMem);
	        UnLockSharedMemory(LockFileDesc);
            //if(debug) syslog(LOG_CRIT, "RegisterForEvents() unlocked shared memory.\n");
            return RC_ALREADY_REGISTERED;
        }
	    else
	    {
	        pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting       = TRUE;
	        pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived = FALSE;
        }
    }

    // LOOP FOREVER WITH SHARED MEMORY LOCKED, SHARED MEMORY MUST BE
    // UNLOCKED ON EXIT
    while (TRUE)
    {
       IgnoreEvent = TRUE;

       // Start of PHR_225158
       // CHECK TO SEE IF AN INTERNALLY GENERATED EVENT WAS MISSED SINCE IT'S LAST REGISTRATION
       if ( (pSpSharedMem->ResetEventInfo.event) &&
            (pSpSharedMem->ResetEventInfo.node == Node))
       {
           TempPtr = EventBuff;
           InitResetEventBuff(pSpSharedMem, TempPtr);

           memcpy(CmdBlock, TempPtr, 15);

           pSpSharedMem->ResetEventInfo.event =  0;
           pSpSharedMem->ResetEventInfo.node  = -1;

           pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting       = FALSE;
           pSpSharedMem->CallerList[Handle].EventRegInfo.Deregister    = FALSE;
           pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived = FALSE;

           shmdt(SharedMem);
           UnLockSharedMemory(LockFileDesc);
           //if(debug) syslog(LOG_CRIT, "RegisterForEvents() unlocked shared memory.\n");
           return RC_SUCCESS;
       }
       // End of PHR_225158

       // CHECK TO SEE IF THIS PROCESS MISSED AN EVENT SINCE ITS LAST REGISTRATION
       if (pSpSharedMem->SPDevNode[Node].EventIndex != pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex)
       {
           //if(debug) syslog(LOG_CRIT, "RegForEvents() process missed an event, returning missed event.\n");
           pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived = TRUE;
       }

       EventIndex = pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex;
       EvtCmdLen  = (int)(pSpSharedMem->SPDevNode[Node].EventQueue.Event[EventIndex])[1];
       TempPtr    = pSpSharedMem->SPDevNode[Node].EventQueue.Event[EventIndex];
       TempPtr   += 6;

       //printf("ASMLIB: Global SP nodes event index = %i and callers index = %i.\n",
       //        pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex,
       //        pSpSharedMem->SPDevNode[node].EventIndex);

       //printf("ASMLIB: Incremented callers event index, index = %i.\n",
       //pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex);

       // ADJUST THE EVENT LISTS EVENT INDEX
       pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex += 1;

       if (pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex == EVENT_QUEUE_DEPTH)
       {
           pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex = 0;
       }

       if(!pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived)
       {

           // UNLOCK SHARED MEMORY WHILE WAITING FOR AN EVENT TO OCCUR
           UnLockSharedMemory(LockFileDesc);
           //if(debug) syslog(LOG_CRIT, "RegisterForEvents() unlocked shared memory.\n");

           // WAIT FOR AN EVENT TO OCCUR
           while(!pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived)
           {
              if (pSpSharedMem->CallerList[Handle].EventRegInfo.Deregister)
              {
                  break;
              }

              // Start of PHR_225158
              // CHECK TO SEE IF AN INTERNALLY GENERATED EVENT HAS TO BE REPORTED
              if ( (pSpSharedMem->ResetEventInfo.event) &&
                   (pSpSharedMem->ResetEventInfo.node == Node))
              {
                  LockFileDesc = LockSharedMemory();

                  if (LockFileDesc == -1)
                  {
                      //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Could not get access to shared memory, Returning RC_OPEN_FAILED.\n");
                      shmdt(SharedMem);
                      return RC_SYSTEM_ERROR;
                  }

                  //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Locked shared memory.\n");
                  TempPtr = EventBuff;
                  InitResetEventBuff(pSpSharedMem, TempPtr);

                  memcpy(CmdBlock, TempPtr, 15);

                  pSpSharedMem->ResetEventInfo.event =  0;
                  pSpSharedMem->ResetEventInfo.node  = -1;

                  pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting       = FALSE;
                  pSpSharedMem->CallerList[Handle].EventRegInfo.Deregister    = FALSE;
                  pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived = FALSE;

                  shmdt(SharedMem);
                  UnLockSharedMemory(LockFileDesc);
                  //if(debug) syslog(LOG_CRIT, "RegisterForEvents() unlocked shared memory.\n");
                  return RC_SUCCESS;
              }
              // End of PHR_225158

              sleep(3);
           }

	       LockFileDesc = LockSharedMemory();

	       if (LockFileDesc == -1)
	       {
	           //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Could not get access to shared memory, Returning RC_OPEN_FAILED.\n");
	           shmdt(SharedMem);
	           return RC_SYSTEM_ERROR;
	       }

	       //if(debug) syslog(LOG_CRIT, "RegisterForEvents() Locked shared memory.\n");
       }

       if (pSpSharedMem->CallerList[Handle].EventRegInfo.Deregister)
       {
           rc = RC_EVENT_CANCELLED;
           break;
       }

       // DETERMINE WHETHER TO IGNORE THIS EVENT
       for (i = 0, IgnoreEvent=FALSE; i < count; i++)
       {
           // SKIP A NULL ARRAY
           if (CmdArray[i] == NULL) continue;

           // IF THE STRINGS MATCH, THEN WE'RE IGNORING THIS EVENT - STOP PARSING IGNORE LIST
           rc = strncmp( CmdArray[i], TempPtr, EvtCmdLen);

           if (rc == 0)
           {
               IgnoreEvent = TRUE;
               break;
           }
       }

       if (IgnoreEvent)
       {
           //if (EventReceived)
           //{
           //   printf("EVTREG: Received a %X.%X.%X event that is being ignored\n",
           //           TempPtr[0], TempPtr[1], TempPtr[2]);
           //}

           pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived = FALSE;
       }
       else
       {
           rc = RC_SUCCESS;
           pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting = FALSE;

           // GET THE LENGTH OF THE EVENT DATA
           BuffLen =   6                                                 +	      // Response Header Length
           pSpSharedMem->SPDevNode[Node].EventQueue.Event[EventIndex][1] +        // cmd length
           pSpSharedMem->SPDevNode[Node].EventQueue.Event[EventIndex][2] +        // data length LSB
          (pSpSharedMem->SPDevNode[Node].EventQueue.Event[EventIndex][3] * 256);  // data length MSB


           // MAKE SURE IT FITS INTO OUR EVENT BUFFER
           if (BuffLen >= pSpSharedMem->CallerList[Handle].BufferSize)
           {
               //if(debug) syslog(LOG_CRIT, "Event contains more data than callers buffer can hold, truncating data.\n");
	           pSpSharedMem->CallerList[Handle].ReturnCode = RC_INVALID_BUFFER;
	           break;
           }

           memset(CmdBlock, 0, BuffLen);

           memcpy(CmdBlock,
                  pSpSharedMem->SPDevNode[Node].EventQueue.Event[EventIndex],
	              BuffLen);

           break;
       }
   }   // end while (!InterruptedBySignal);

   pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting       = FALSE;
   pSpSharedMem->CallerList[Handle].EventRegInfo.Deregister    = FALSE;
   pSpSharedMem->CallerList[Handle].EventRegInfo.EventReceived = FALSE;

   // UNLOCK SHARED MEMORY
   shmdt(SharedMem);
   UnLockSharedMemory(LockFileDesc);
   //if(debug) syslog(LOG_CRIT, "RegisterForEvents() unlocked shared memory.\n");

   return rc;
}


unsigned int DeregisterForEvents(SYS_HANDLE Handle, unsigned long Reserved)
{
   int   LockFileDesc;
   int   MemID;
   void *SharedMem;
   int   Timeout;

   PSP_SHARED_MEM pSpSharedMem;

   // VALIDATE CALLERS PARMS
   if ( (Handle <= 0)  || (Handle > MAX_HANDLES) )
   {
      //printf("ASMLIB: DeRegisterForEvents() - Invalid Handle detected, returning RC_INVALID_HANDLE.\n");
       return RC_INVALID_HANDLE;
   }

   // ALLOCATE SHARED MEMORY
   pSpSharedMem = NULL;

   MemID = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

   //if(debug) syslog(LOG_CRIT, "AllocateSharedMem() MemID = %i. \n", MemID);

   if ( MemID == IPC_FAILURE )
   {
        //if(debug) syslog(LOG_CRIT, " shmget() failed, returning RC_SYSTEM_ERROR.\n");
        return RC_SYSTEM_ERROR;
   }

   // ATTACH THIS PROCESS TO THE SHARED MEMORY
   SharedMem = shmat(MemID, (void *)0, 0);

   if (SharedMem == (void *)IPC_FAILURE )
   {
        //if(debug) syslog(LOG_CRIT, " shmat() failed, returning RC_SYSTEM_ERROR.\n");
        return RC_SYSTEM_ERROR;
   }

   // SET UP THE LOCAL POINTER TO SHARED MEMORY
   pSpSharedMem = (PSP_SHARED_MEM)SharedMem;

   // MAKE SURE THAT THE SP IS NOT CURRENTLY RESETTING AND CHECK
   // FOR AT LEAST 1 VALID SP IN THE SYSTEM
   Timeout = 0;

   while ((pSpSharedMem->SpCount == 0) || (pSpSharedMem->Resetting))
   {
	   sleep(1);

	   Timeout++;

	   if (Timeout >= CMD_TIMEOUT)
	   {
		   //if (pSpSharedMem->SpCount == 0)
		   //{
				//if(debug) syslog(LOG_CRIT, "DeRegisterForEvents() Could not find a valid SP device in the system, returning RC_TIMEOUT.\n");
		   //}
		   //else
		   //{
				//if(debug) syslog(LOG_CRIT, "DeRegisterForEvents() Detected SP reset in progress, returning RC_TIMEOUT.\n");
		   //}

		   shmdt(SharedMem);
		   return RC_TIMEOUT;
	   }
   }

   LockFileDesc = LockSharedMemory();

   if (LockFileDesc == -1)
   {
       //printf("ASMLIB: Could not get access to shared memory, Returning RC_OPEN_FAILED.\n");
       shmdt(SharedMem);
       return RC_SYSTEM_ERROR;
   }

   //if(debug) syslog(LOG_CRIT, "DeRegisterForEvents() Locked shared memory.\n");

   // ENSURE HANDLE IS VALID
   if (pSpSharedMem->HandleList[Handle].Free)
   {
       // CLEAN UP AND EXIT
       //printf("ASMLIB: DeRegisterForEvents() - Invalid Handle detected, returning RC_INVALID_HANDLE.\n");
	   shmdt(SharedMem);
       UnLockSharedMemory(LockFileDesc);
       //if(debug) syslog(LOG_CRIT, "DeRegisterForEvents() unlocked shared memory.\n");
       return RC_INVALID_HANDLE;
   }

   // ENSURE THIS PROCESS IS CURRENTLY REGISTERED FOR EVENTS
   if (pSpSharedMem->CallerList[Handle].EventRegInfo.Initialized == FALSE )            // PHR_EVENT
   {
       // CLEAN UP AND EXIT
       //printf("ASMLIB: DeRegisterForEvents() - Process is not registered for events, returning RC_EVENT_NOT_REGISTERED.\n");
	   shmdt(SharedMem);
       UnLockSharedMemory(LockFileDesc);
       //if(debug) syslog(LOG_CRIT, "DeRegisterForEvents() unlocked shared memory.\n");
       return RC_EVENT_NOT_REGISTERED;
   }

   pSpSharedMem->CallerList[Handle].EventRegInfo.Deregister = TRUE;

   // INFORM THE REGISTERED/WAITING TASK THAT IT IS GETTING DE-REGEISTERED
   // kill( pSpSharedMem->CallerList[Handle].EventRegInfo.Pid, SIGINT );

   // UPDATE SHARED MEMORY BY REMOVING THIS PROCESS FROM THE EVENT LIST
   pSpSharedMem->CallerList[Handle].EventRegInfo.Pid         = -1;
   pSpSharedMem->CallerList[Handle].EventRegInfo.Handle      = -1;
   pSpSharedMem->CallerList[Handle].EventRegInfo.Node        = -1;
   pSpSharedMem->CallerList[Handle].EventRegInfo.EventIndex  = -1;

   pSpSharedMem->CallerList[Handle].EventRegInfo.Deregister  = TRUE;
   pSpSharedMem->CallerList[Handle].EventRegInfo.Waiting     = FALSE;   // PHR_EVENT
   pSpSharedMem->CallerList[Handle].EventRegInfo.Initialized = FALSE;   // PHR_EVENT

   shmdt(SharedMem);
   UnLockSharedMemory(LockFileDesc);
   //if(debug) syslog(LOG_CRIT, "DeRegisterForEvents() unlocked shared memory.\n");
   return RC_SUCCESS;
}


////////////////////////////////////////////////////////////////////////
//
// RegisterForReverseHB
//
// Starts a periodic check to see if the Service Processor is alive.
// This routine uses the ReverseHeartBeatRunning flag to see if this
// test has already been started, or if the test should be stopped.
//
// This function returns to the calling process only when the Service
// Processor fails to respond to a simple periodic command, or if the
// ReverseHeartBeatRunning flag has been set to FALSE by the
// DeRegisterForReverseHB() function.
//
// Possible return codes:
//     RC_SYSTEM_ERROR
//     RC_REVERSE_HB_ALREADY_REGISTERD
//     RC_REVERSE_HEARTBEAT_HAS_FAILED
//     RC_SUCCESS
//
////////////////////////////////////////////////////////////////////////
unsigned int RegisterForReverseHB(SYS_HANDLE Handle, unsigned long reserved)
{
   unsigned int   node;
   unsigned char  Buffer[1024];
   int            rc;
   int            LockFileDesc;
   int            Timeout;

   // SHARED MEMORY DECLARATIONS
   int   MemID;
   void *SharedMem;

   PSP_SHARED_MEM pSpSharedMem;


   // VALIDATE CALLERS PARMS
   if ( (Handle <= 0)  || (Handle > MAX_HANDLES) )
   {
       //printf("ASMLIB: DeRegisterForEvents() - Invalid Handle detected, returning RC_INVALID_HANDLE.\n");
       return RC_INVALID_HANDLE;
   }

   // ALLOCATE SHARED MEMORY
   pSpSharedMem = NULL;

   MemID = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

   //if(debug) syslog(LOG_CRIT, "AllocateSharedMem() MemID = %i. \n", MemID);

   if ( MemID == IPC_FAILURE )
   {
        //if(debug) syslog(LOG_CRIT, " shmget() failed, returning RC_SYSTEM_ERROR.\n");
        return RC_SYSTEM_ERROR;
   }

   // ATTACH THIS PROCESS TO THE SHARED MEMORY
   SharedMem = shmat(MemID, (void *)0, 0);

   if (SharedMem == (void *)IPC_FAILURE )
   {
        //if(debug) syslog(LOG_CRIT, " shmat() failed, returning RC_SYSTEM_ERROR.\n");
        return RC_SYSTEM_ERROR;
   }

   // SET UP THE LOCAL POINTER TO SHARED MEMORY
   pSpSharedMem = (PSP_SHARED_MEM)SharedMem;

   // MAKE SURE THAT THE SP IS NOT CURRENTLY RESETTING AND CHECK
   // FOR AT LEAST 1 VALID SP IN THE SYSTEM
   Timeout = 0;

   while ((pSpSharedMem->SpCount == 0) || (pSpSharedMem->Resetting))
   {
	   sleep(1);

	   Timeout++;

	   if (Timeout >= CMD_TIMEOUT)
	   {
		   //if (pSpSharedMem->SpCount == 0)
		   //{
				//if(debug) syslog(LOG_CRIT, "RegisterForReverseHB() Could not find a valid SP device in the system, returning RC_TIMEOUT.\n");
		   //}
		   //else
		   //{
				//if(debug) syslog(LOG_CRIT, "RegisterForReverseHB() Detected SP reset in progress, returning RC_TIMEOUT.\n");
		   //}

		   shmdt(SharedMem);
		   return RC_TIMEOUT;
	   }
   }

   LockFileDesc = LockSharedMemory();

   if (LockFileDesc == -1)
   {
       //printf("ASMLIB: Could not get access to shared memory, Returning RC_OPEN_FAILED.\n");
       shmdt(SharedMem);
       return RC_TIMEOUT;
   }

   //if(debug) syslog(LOG_CRIT, "RegisterForRHB() Locked shared memory.\n");

   node = GetNodeFromHandle(pSpSharedMem, Handle);

   // ENSURE THE CALLERS NODE IS VALID
   if ( (node >= MAX_SP_NODES) || (node < 0) )
   {
       //printf("ASMLIB: detected an invalid node, returning RC_OPEN_FAILED.\n");
	   shmdt(SharedMem);
       UnLockSharedMemory(LockFileDesc);
       //if(debug) syslog(LOG_CRIT, "RegisterForRHB() unlocked shared memory.\n");
       return RC_INVALID_HANDLE;
   }

   // ENSURE THE NODE IS OPEN AND ACCESSABLE
   if(pSpSharedMem->SPDevNode[node].State != NODE_OPENED)
   {
       //printf("ASMLIB: detected an invalid node, returning RC_OPEN_FAILED.\n");
	   shmdt(SharedMem);
       UnLockSharedMemory(LockFileDesc);
       //if(debug) syslog(LOG_CRIT, "RegisterForRHB() unlocked shared memory.\n");
       return RC_SYSTEM_ERROR;
   }

   // ******************************************************************
   // IF THE CALLER HAS NOT ALREADY REGISTERED FOR REVERSE HEARTBEAT
   // NOTIFICATION, THEN SET THE ReverseHeartBeatRunning FLAG TO TRUE,
   // OTHERWISE IT IS ALREADY REGISTERED.
   // ******************************************************************
   if (pSpSharedMem->CallerList[Handle].RegForRH == TRUE)
   {
       //if(debug) syslog(LOG_CRIT, "RegsiterForReverseHB() detected PID %i is already registered, returning RC_ALREADY_REGISTERD.\n",
       //       pSpSharedMem->CallerList[Handle].Pid);
	   shmdt(SharedMem);
       UnLockSharedMemory(LockFileDesc);
       //if(debug) syslog(LOG_CRIT, "RegisterForRHB() unlocked shared memory.\n");
       return  RC_REVERSE_HB_ALREADY_REGISTERD;
   }
   else   // else this process can regiseter for RHB
   {
       //if(debug) syslog(LOG_CRIT, "RegsiterForReverseHB() PID %i is registering for RHB, handle = %i.\n",
       //       pSpSharedMem->CallerList[Handle].Pid, Handle);

       pSpSharedMem->CallerList[Handle].RegForRH = TRUE;
   }

   // The call to SystemDataIO will lock shared memory so unlock it here
   UnLockSharedMemory(LockFileDesc);
   //if(debug) syslog(LOG_CRIT, "RegisterForRHB() unlocked shared memory.\n");

   while(pSpSharedMem->CallerList[Handle].RegForRH)
   {
       memset(Buffer, 0, 1024);

       // PREPARE GET SYSTEM STATE COMMAND BUFFER
       Buffer[0] = 0x02;          // Read
       Buffer[1] = 0x03;          // Command length
       Buffer[2] = 0x00;          // Data length LSB
       Buffer[3] = 0x00;          // Data Length MSB
       Buffer[4] = 0x00;          // Status
       Buffer[5] = 0x00;          // Reserved
       Buffer[6] = 0x04;          // 1st cmd byte - System State Command
       Buffer[7] = 0x03;          // 2nd cmd byte
       Buffer[8] = 0x06;          // 3rd cmd byte

	   rc = SystemDataIO(Handle, &Buffer, 0);

       if(!pSpSharedMem->CallerList[Handle].RegForRH)
	   {
	       rc = RC_SUCCESS;
	       break;
	   }

       if ( (Buffer[4] != 0) || (rc != RC_SUCCESS) )
	   {
	       rc = RC_REVERSE_HEARTBEAT_HAS_FAILED;
	       break;
	   }

       sleep(5);
   }

   if (!pSpSharedMem->CallerList[Handle].RegForRH)
   {
       //if(debug) syslog(LOG_CRIT, "RegsiterForReverseHB() PID %i has DeRegisteredForRevHB, returning RC_SUCCESS.\n",
       //     pSpSharedMem->CallerList[Handle].Pid);

       rc = RC_SUCCESS;
   }
   else if (rc != RC_SUCCESS)
   {
       //if(debug) syslog(LOG_CRIT, "RegsiterForReverseHB() detected RHB failure, returning to caller.\n");
       rc = RC_REVERSE_HEARTBEAT_HAS_FAILED;
   }

   shmdt(SharedMem);
   return rc;
}



unsigned int DeRegisterForReverseHB(SYS_HANDLE Handle, unsigned long reserved)
{
   int rc = RC_SUCCESS;
   int LockFileDesc;
   int Timeout;

   // SHARED MEMORY DECLARATIONS
   int   MemID;
   void *SharedMem;

   PSP_SHARED_MEM pSpSharedMem;

   // VALIDATE CALLERS PARMS
   if ( (Handle <= 0)  || (Handle > MAX_HANDLES) )
   {
       //printf("ASMLIB: DeRegisterForEvents() - Invalid Handle detected, returning RC_INVALID_HANDLE.\n");
       return RC_INVALID_HANDLE;
   }

   // ALLOCATE SHARED MEMORY
   pSpSharedMem = NULL;

   MemID = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

   //if(debug) syslog(LOG_CRIT, "AllocateSharedMem() MemID = %i. \n", MemID);

   if ( MemID == IPC_FAILURE )
   {
        //if(debug) syslog(LOG_CRIT, " shmget() failed, returning RC_SYSTEM_ERROR.\n");
        return RC_SYSTEM_ERROR;
   }

   // ATTACH THIS PROCESS TO THE SHARED MEMORY
   SharedMem = shmat(MemID, (void *)0, 0);

   if (SharedMem == (void *)IPC_FAILURE )
   {
        //if(debug) syslog(LOG_CRIT, " shmat() failed, returning RC_SYSTEM_ERROR.\n");
        return RC_SYSTEM_ERROR;
   }

   // SET UP THE LOCAL POINTER TO SHARED MEMORY
   pSpSharedMem = (PSP_SHARED_MEM)SharedMem;

   // MAKE SURE THAT THE SP IS NOT CURRENTLY RESETTING AND CHECK
   // FOR AT LEAST 1 VALID SP IN THE SYSTEM
   Timeout = 0;

   while ((pSpSharedMem->SpCount == 0) || (pSpSharedMem->Resetting))
   {
	   sleep(1);

	   Timeout++;

	   if (Timeout >= CMD_TIMEOUT)
	   {
		   //if (pSpSharedMem->SpCount == 0)
		   //{
				//if(debug) syslog(LOG_CRIT, "DeRegisterForReverseHB() Could not find a valid SP device in the system, returning RC_TIMEOUT.\n");
		   //}
		   //else
		   //{
				//if(debug) syslog(LOG_CRIT, "DeRegisterForReverseHB() Detected SP reset in progress, returning RC_TIMEOUT.\n");
		   //}

		   shmdt(SharedMem);
		   return RC_TIMEOUT;
	   }
   }


   LockFileDesc = LockSharedMemory();

   if (LockFileDesc == -1)
   {
       //printf("ASMLIB: Could not get access to shared memory, Returning RC_OPEN_FAILED.\n");
       shmdt(SharedMem);
       return RC_TIMEOUT;
   }

   //if(debug) syslog(LOG_CRIT, "DeRegisterForRHB() Locked shared memory.\n");

   // ******************************************************************
   // IF THE CALLER HAS NOT ALREADY REGISTERED FOR REVERSE HEARTBEAT
   // NOTIFICATION, THEN SET THE ReverseHeartBeatRunning FLAG TO TRUE,
   // OTHERWISE IT IS ALREADY REGISTERED.
   // ******************************************************************
   if (pSpSharedMem->CallerList[Handle].RegForRH == TRUE)
   {
       //if(debug) syslog(LOG_CRIT, "DeRegsiterForReverseHB() PID %i is De-registering for RHB, handle = %i.\n",
       //       pSpSharedMem->CallerList[Handle].Pid, Handle);

       pSpSharedMem->CallerList[Handle].RegForRH = FALSE;
       rc = RC_SUCCESS;
   }
   else   // else this handle has not regisetered for RHB
   {
       //if(debug) syslog(LOG_CRIT, "DeRegsiterReverseHB() PID %i not registered, returning RC_REVERSE_HB_ALREADY_STOPPED.\n",
       //       pSpSharedMem->CallerList[Handle].Pid);

       rc = RC_REVERSE_HB_ALREADY_STOPPED;
   }

   shmdt(SharedMem);
   UnLockSharedMemory(LockFileDesc);

   //if(debug) syslog(LOG_CRIT, "DeRegisterForRHB() unlocked shared memory.\n");
   return rc;
}





// ************************************************************************
// * CloseSPDriver - This routine must be called by an application that   *
// *                 previously called OpenSPDriver or OpenSPDriverNode   *
// *                 It issues a usb_close and deregisters the caller     *
// *                 from the global SpDeviceNode                         *
// *                                                                      *
// * Valid Return Codes to Application :                                  *
// *      RC_INVALID_HANDLE       - invalid handle                        *
// *      RC_CLOSE_FAILED         - failed to close                       *
// *      RC_SUCCESS              - close was successful                  *
// ************************************************************************
unsigned int CloseSPDriver(SYS_HANDLE Handle, unsigned long  Reserved)
{
    int rc;
	int Timeout;
    int CallersPid;
    int LockFileDesc;

    int SharedMemId;
    PSP_SHARED_MEM pSpSharedMem;
    void *SharedMem;

    // DISPLAY WELCOME MESSAGE
     if(debug) syslog(LOG_CRIT, "Entered CloseSPDriver - closing handle %i.\n", Handle);

    // ALLOCATE THE SHARED MEMORY THAT THE CALLING PROCESSES AND THIS
    // PROCESS NEED ACCESS TO
    SharedMemId = shmget( (key_t)SHARED_MEM_KEY, sizeof(SP_SHARED_MEM), 0666 | IPC_CREAT);

    if (SharedMemId == IPC_FAILURE )
    {
        if(debug) syslog(LOG_CRIT, "CloseSPDriver: shmget() failed returning failure.\n");
        return RC_SYSTEM_ERROR;
    }

    if(debug) syslog(LOG_CRIT, "CloseSPDriver: shmget() succeded, SharedMemId = %X.\n", SharedMemId);

    // ATTACH THIS PROCESS TO THE SHARED MEMORY
    SharedMem = shmat(SharedMemId, (void *)0, 0);

    if (SharedMem == (void *)IPC_FAILURE )
    {
        if(debug) syslog(LOG_CRIT, "CloseSPDriver: shmat() failed returning failure.\n");
        return RC_SYSTEM_ERROR;
    }

    if(debug) syslog(LOG_CRIT, "CloseSPDriver: shmat() succeded, attached memory at %X.\n", (int) SharedMem);

    pSpSharedMem = (PSP_SHARED_MEM)SharedMem;

    // MAKE SURE THAT THE SP IS NOT CURRENTLY RESETTING AND CHECK
	// FOR AT LEAST 1 VALID SP IN THE SYSTEM
	Timeout = 0;

	while ((pSpSharedMem->SpCount == 0) || (pSpSharedMem->Resetting))
	{
		sleep(1);

		Timeout++;

		if (Timeout >= CMD_TIMEOUT)
		{
			if (pSpSharedMem->SpCount == 0)
			{
				 if(debug) syslog(LOG_CRIT, "CloseSpDriver() Could not find a valid SP device in the system, returning RC_TIMEOUT.\n");
			}
			else
			{
			     if(debug) syslog(LOG_CRIT, "CloseSpDriver() Detected SP reset in progress, returning RC_TIMEOUT.\n");
			}

			shmdt(SharedMem);
			return RC_CLOSE_FAILED;
		}
	}


    LockFileDesc = LockSharedMemory();

    if (LockFileDesc == -1)
    {
        if(debug) syslog(LOG_CRIT, "CloseSPDriver: Could not get access to shared memory, Returning RC_OPEN_FAILED.\n");
        shmdt(SharedMem);
        return RC_CLOSE_FAILED;
    }

    if(debug) syslog(LOG_CRIT, "CloseSpDriver() Locked shared memory.\n");

    CallersPid = getpid();

    rc = RemoveCaller(pSpSharedMem, Handle, CallersPid);

    if ( rc != RC_SUCCESS )
    {
        if(debug) syslog(LOG_CRIT, "CloseSPDriver: Invalid handle passed by caller, returning RC_INVALID_HANDLE.\n");
        rc = RC_INVALID_HANDLE;
    }
    else
    {
	    rc = FreeHandle(pSpSharedMem,Handle);

	    if ( rc != RC_SUCCESS )
	    {
	        if(debug) syslog(LOG_CRIT, "CloseSPDriver: Invalid handle passed by caller, returning RC_INVALID_HANDLE.\n");
	        rc = RC_INVALID_HANDLE;
	    }
    }

	shmdt(SharedMem);
    UnLockSharedMemory(LockFileDesc);
    if(debug) syslog(LOG_CRIT, "CloseSpDriver() unlocked shared memory.\n");
    if(debug) syslog(LOG_CRIT, "Exiting CloseSpDriver\n");
    return rc;
}



int GetHandle(PSP_SHARED_MEM pSpSharedMem)
{
   int i;
   int handle = INVALID_USB_HANDLE;

   for ( i = 1; i <= MAX_HANDLES; i++)
   {
      if ( pSpSharedMem->HandleList[i].Free )
      {
          pSpSharedMem->HandleList[i].Free = FALSE;
          handle = i;
          break;
      }
   }

   return handle;
}





int FreeHandle(PSP_SHARED_MEM pSpSharedMem, int Handle)
{

    if ( (Handle <= 0) || (Handle > MAX_HANDLES) )
    {
        return INVALID_USB_HANDLE;
    }

    pSpSharedMem->HandleList[Handle].Free = TRUE;

    return RC_SUCCESS;
}



// ********************************************************************
// * AddCaller - This routine returns a free handle from the list of  *
// *             handles, if there are no free handles it returns a   *
// *             -1 (error).                                          *
// ********************************************************************
int AddCaller( PSP_SHARED_MEM  pSpSharedMem,
               int             CallersPid,
               unsigned int    BuffSize,
               unsigned int    Node,
               int             Handle )
{
   if ( (Handle <= 0) || (Handle > MAX_HANDLES) )
   {
       return INVALID_USB_HANDLE;
   }

   pSpSharedMem->CallerList[Handle].ReturnCode = -1;
   pSpSharedMem->CallerList[Handle].Pid        = CallersPid;
   pSpSharedMem->CallerList[Handle].BufferSize = BuffSize * 1024;
   pSpSharedMem->CallerList[Handle].Node       = Node;

   return RC_SUCCESS;
}

// ********************************************************************
// * FreeHandle - This routine frees the handle, returns -1 if fails  *
// ********************************************************************
int RemoveCaller(PSP_SHARED_MEM pSpSharedMem, int handle, int CallersPid)
{
   int rc;

   if ( (handle <= 0) || (handle > MAX_HANDLES) )
   {
       return INVALID_USB_HANDLE;
   }

   if (pSpSharedMem->CallerList[handle].Pid == CallersPid)
   {
       pSpSharedMem->CallerList[handle].Node  = -1;
       pSpSharedMem->CallerList[handle].Pid   = -1;

       rc = RC_SUCCESS;
   }
   else
   {
       rc = INVALID_USB_HANDLE;
   }

   return rc;
}

unsigned char  usb_cksum( unsigned char* buf, int size)
{
   unsigned char sum;

   sum = 0;

   while (size > 0)
   {
      sum += *(unsigned short*)buf;        // add the next word to "sum"
      buf += 1;                            // point to next word in slim packet
      size -= 1;                           // dec "size" by 2 bytes
   }

   sum = 1 + ~sum;
   return sum;
}


void UnLockSharedMemory(int FileDesc)
{
    //if(debug) syslog(LOG_CRIT, "Unlocking File descriptor %i.\n",FileDesc);
    close(FileDesc);
    unlink(MEM_LOCK_FILE);
}

int GetNodeFromHandle(PSP_SHARED_MEM pSpSharedMem, SYS_HANDLE Handle)
{
     return pSpSharedMem->CallerList[Handle].Node;
}


int ValidateCaller( PSP_SHARED_MEM pSpSharedMem, int Handle, int CallersPid)
{
    //int rc;

    if (pSpSharedMem->HandleList[Handle].Free)
    {
        //printf("ASMLIB: ValidateCaller() Handle not free.\n");
        return RC_INVALID_HANDLE;
    }

    pSpSharedMem->CallersNode   = pSpSharedMem->CallerList[Handle].Node;
    pSpSharedMem->CallersHandle = Handle;

    return RC_SUCCESS;
}



//**************************************************************************************
// USE THE FOLLOWING LOCK FILE METHOD TO MAKE SURE THAT IBMASM IS NOT ALREADY LOADED   *
// TRY TO OPEN THE /tmp/ibmasm.pid FILE WITH READ ACCESS. IF IT OPENS SUCCESSFULLY     *
// IBMASM IS ALREADY LOAEDED SO CLOSE THE FILE AND ABORT. OTHERWISE IT WAS NOT CREATED *
// SO OPEN THE FILE WITH WRITE ACCESS, STORE THE PID AND CONTINUE                      *
//**************************************************************************************
int LockSharedMemory(void)
{
    int FileDesc;
    int TimeOut = 0;

    while(TimeOut < CMD_TIMEOUT)
    {
	    FileDesc = -1;

	    FileDesc = open(MEM_LOCK_FILE, O_CREAT | O_EXCL);

        if (FileDesc == -1)
        {
	       TimeOut += 1;

           if (TimeOut >= CMD_TIMEOUT)
           {
               //if(debug) syslog(LOG_CRIT, "open of MEM_LOCK_FILE failed with a timeout.\n");
	          return -1;
	       }

	       sleep(1);
	    }
	    else
	    {
	       break;
	    }
    }

    //if(debug) syslog(LOG_CRIT, "locking with file descriptor %i.\n",FileDesc);

    return FileDesc;
}


              // Start of PHR_225158
void InitResetEventBuff(PSP_SHARED_MEM pSpSharedMem, unsigned char *EventBuff)
{
    int event;

    event = pSpSharedMem->ResetEventInfo.event;

    EventBuff[0] = 5;                        // Command Type
    EventBuff[1] = 3;                        // Command Length
    EventBuff[2] = 1;                        // Data Length  LSB
    EventBuff[3] = 0;                        // Data Length  MSB
    EventBuff[4] = 0;                        // Status
    EventBuff[5] = 0;                        // Reserved
    EventBuff[6] = 5;                        // Command Byte 0
    EventBuff[7] = 2;                        // Command Byte 1
    EventBuff[8] = 16;                       // Command Byte 2
    EventBuff[9] = event;                    // Data Byte 0

    /*switch (event)
    {
       case EVENT_RESET_COMPLETE:
          if(debug) syslog(LOG_CRIT, "RegForEvents() is reporting a EVENT_RESET_COMPLETE.\n");
          break;
       case EVENT_DISCONNECT:
          if(debug) syslog(LOG_CRIT, "RegForEvents() is reporting a EVENT_DISCONNECT.\n");
          break;
       case EVENT_RECONNECT:
          if(debug) syslog(LOG_CRIT, "RegForEvents() is reporting a EVENT_RECONNECT.\n");
          break;
    }
    */

    return;
}
// End of PHR_225158

